/*
 * This file declares some parameters which is common to all IPC mechanisms. Every IPC mechanism
 * inherits this class and implements the functions declared. Since Java has runtime binding
 * so the corresponding methods will be called.
 * 
 * MAXNUMTHREADS - The maximum number of java threads running
 * EMUTHREADS - The number of emulator threads 1 java thread is reading from
 * COUNT - this many number of packets is allocated in the shared memory for each 
 * 		   application/emulator thread 
 * */

package emulatorinterface.communication;

import java.util.concurrent.Semaphore;
import config.SystemConfig;

import emulatorinterface.*;
import generic.CircularPacketQueue;

public abstract class IpcBase {

	protected static int MaxNumJavaThreads = SystemConfig.maxNumJavaThreads;
	protected static int EmuThreadsPerJavaThread = SystemConfig.numEmuThreadsPerJavaThread; 
	protected static int MaxNumThreads = MaxNumJavaThreads*EmuThreadsPerJavaThread;
	
	// state management for reader threads
	public boolean[] termination=new boolean[MaxNumJavaThreads];
	public boolean[] started=new boolean[MaxNumJavaThreads];

	// number of instructions read by each of the threads
	public long[] numInstructions = new long[MaxNumJavaThreads];

	// to maintain synchronization between main thread and the reader threads
	public static final Semaphore free = new Semaphore(0, true);

	public static GlobalTable glTable;

	StreamGobbler s1;
	StreamGobbler s2;
	
	// Initialise structures and objects
	public IpcBase () {
		glTable = new GlobalTable(this);
	}

	// must be same as in IPCBase.cc
	public static int getEmuId(int tidApp) {
		if (tidApp>=MaxNumThreads) misc.Error.shutDown("tidApp more than supported "+tidApp);
		return tidApp/MaxNumJavaThreads;
	}
	public static int getJavaId(int tidApp) {
		if (tidApp>=MaxNumThreads) 
			
		{
			System.out.println("tidApp more than supported "+tidApp);
			misc.Error.shutDown("tidApp more than supported "+tidApp);
			}
		return tidApp%MaxNumJavaThreads;
	}
	public static int getAppId(int tid, int tidEmu) {
		return tidEmu*MaxNumJavaThreads+tid;
	}

	public abstract boolean errorCheck(int tidApp, long totalReads);
	
	public boolean someThreadStarted() {
		for (int i=0; i<MaxNumJavaThreads; i++) {
			if (started[i]) return true;
		}
		return false;
	}
	
	// Start the PIN process. Parse the cmd accordingly
	public Process startPIN(String cmd) throws Exception {
		Runtime rt = Runtime.getRuntime();
		try {
			Process p = rt.exec(cmd);
			s1 = new StreamGobbler ("stdin", p.getInputStream ());
			s2 = new StreamGobbler ("stderr", p.getErrorStream ());
			s1.start ();
			s2.start ();
			return p;
		} catch (Exception e) {
			e.printStackTrace();
			return null;
		}
	}

	public abstract void initIpc();


	/*** start, finish, isEmpty, fetchPacket, isTerminated ****/
	public RunnableThread[] getRunnableThreads() {
		System.out.println("Implement getRunnableThreads() in the IPC mechanism");
		return null;
	}


	public long doExpectedWaitForSelf() throws InterruptedException{
		
		// this takes care if no thread started yet.
				free.acquire();	
				
				int j=0;

				//hack. do it via fini in pin.
				while (!termination[0]) {
					free.acquire();
					j++;
				}
					
		/*		// if any thread has started and not finished then wait.
				for (int i=0; i<MaxNumJavaThreads; i++) {
					if (started[i] && !termination[i]) {
						free.acquire();
						j++;
					}
				}
		*/		
				//inform threads which have not started about finish
				for (int i=0; i<MaxNumJavaThreads; i++) {
					if (started[i]==false) {
						termination[i]=true;
					}
					//totalInstructions += numInstructions[i];
				}
				System.out.println("Here !!!!");
				for (; j<MaxNumJavaThreads-1; j++)
					free.acquire();
				System.out.println("Here111 !!!!");
				
				//s1.join();
				//s2.join();
				//return totalInstructions;
				
				return 0;
	}
	

	// Should wait for PIN too before calling the finish function to deallocate stuff related to
	// the corresponding mechanism
	public void doWaitForPIN(Process p) throws Exception{
		try {
			p.waitFor();
		} catch (Exception e) {

		}
	}

	// Free buffers, free memory , deallocate any stuff.
	public void finish(){
		System.out.println("Implement finish in the IPC mechanism");
	}

	public abstract int fetchManyPackets(int tidApp ,CircularPacketQueue fromEmulator);
}