package generic;

import java.lang.ref.WeakReference;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.atomic.AtomicIntegerArray;

import main.Main;
import misc.Pair;
import generic.PortType;
import generic.Port;
import generic.Event;

/** 
 * Any Object which is supposed to be shared across ClockDomains must extend this 
 *  If the object could be shared as well as not shared(depending on some type)
 *  for example as in Port. then exclude that object from instances as done for 
 *  the ports which are not parallel.
 */

public abstract class SharedInstances {

	private int numPorts;
	private AtomicIntegerArray[] m_PA;
	private AtomicIntegerArray[] m_PACOPY;
	private final int CYCLESIZE = Main.SynchClockDomainsCycles;

	private static List<WeakReference<SharedInstances>> instances = new ArrayList<WeakReference<SharedInstances>>();
	private static List<Pair<Port,Event>> eventInstances = new ArrayList<Pair<Port,Event>>();
	
	public SharedInstances() {
		init(1);
	}

	public SharedInstances(int noOfPorts, PortType portType, long latencyOfConnectedElement) {
		if (portType == PortType.Parallel) {
			init(noOfPorts);
		}
	}

	public SharedInstances(int noOfPorts) {
		init(noOfPorts);
	}
	
	public void init(int noOfPorts) {
		this.numPorts = noOfPorts;
//		System.out.println("SharedInstance "+this.hashCode());
		instances.add(new WeakReference<SharedInstances>(this));
		m_PA=new AtomicIntegerArray[noOfPorts];
		m_PACOPY=new AtomicIntegerArray[noOfPorts];
		for (int i=0; i<noOfPorts; i++) {
			m_PA[i] = new AtomicIntegerArray(CYCLESIZE);
			m_PACOPY[i] = new AtomicIntegerArray(CYCLESIZE);
		}
	}

	public static void registerEvent(Pair<Port,Event> p) {
		SharedInstances.eventInstances.add(p);
	}
	
	// this function is threadsafe. as called from the phaser.
	public static void cleanUp() {
		SharedInstances p;
		for (int i=0; i<SharedInstances.instances.size(); i++) {
			if ((p = SharedInstances.instances.get(i).get()) != null) {
				p.clearSlotsOfInstance();
			}
		}
		
		// Put all the stored events back to the ports
		Pair<Port,Event> pe;
		for (int i=0; i<SharedInstances.eventInstances.size(); i++) {
			if ((pe = SharedInstances.eventInstances.get(i)) != null) {
				Port port = (Port)pe.getL();
				Event e = (Event)pe.getR();
				if(e.getEventQ().getCore().isPipelineInOrder())
					e.getEventQ().getCore().getExecEngine().getFetchUnitIn().setSleep(false);
				else
					e.getEventQ().getCore().getExecEngine().getFetchUnitOut().setSleep(false);
				port.parallePortPutFromCleanup(e);
			}
		}
		
		// Re-empty the event instances list
		eventInstances.clear();
	}

	private void clearSlotsOfInstance() {
		m_PA = m_PACOPY;
		m_PACOPY=new AtomicIntegerArray[numPorts];
		for (int i=0; i<numPorts; i++) {
			m_PACOPY[i] = new AtomicIntegerArray(CYCLESIZE);
		}
	}

	protected long bookSlotOnX (long index, int x, int occupancy) {
		long requestTime = index;
		index %= CYCLESIZE;
		for (int i=(int)index; i<CYCLESIZE; i++) {
			if (m_PA[x].compareAndSet(i, 0, 1))
				return (long)(i-index)+requestTime;
		}
		
		//Look in the shadow array
		for (int i=0; i<CYCLESIZE; i++) {
			if (m_PACOPY[x].compareAndSet(i, 0, 1))
				return (long)(i-index)+CYCLESIZE+requestTime;
		}
		
		return -1;
	}

	// Returns the cycle in which the process will take place
	protected long bookAnySlot(long index,int occupancy) 
	{
		long requestTime = index;

		index%=CYCLESIZE;

		for(int i=(int)index;i<CYCLESIZE;i++)
		{
			for (int j=0; j<numPorts; j++) {
				if (m_PA[j].compareAndSet(i, 0, 1))
					return (long)(i-index)+requestTime;
			}
		}
		
		for(int i=0;i<CYCLESIZE;i++)
		{
			for (int j=0; j<numPorts; j++) {
				if (m_PACOPY[j].compareAndSet(i, 0, 1))
					return (long)(i-index)+CYCLESIZE+requestTime;
			}
		}
		
		return -1;
	}

}
