/*****************************************************************************
				Tejas Simulator
------------------------------------------------------------------------------------------------------------

   Copyright [2010] [Indian Institute of Technology, Delhi]
   Licensed under the Apache License, Version 2.0 (the "License");
   you may not use this file except in compliance with the License.
   You may obtain a copy of the License at

       http://www.apache.org/licenses/LICENSE-2.0

   Unless required by applicable law or agreed to in writing, software
   distributed under the License is distributed on an "AS IS" BASIS,
   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
   See the License for the specific language governing permissions and
   limitations under the License.
------------------------------------------------------------------------------------------------------------

	Contributors:  Prathmesh, Moksh Upadhyay, Mayur Harne
*****************************************************************************/
package memorysystem;

import generic.CommunicationInterface;
import java.util.HashMap;
import java.util.Map;

import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;

import net.BusInterface;
import net.ID;
import net.NocInterface;

import main.ArchitecturalComponent;
import memorysystem.coherence.Coherence;
import memorysystem.coherence.Directory;
import memorysystem.nuca.NucaCache;
import memorysystem.nuca.NucaCache.NucaType;
import config.CacheConfig;
import config.SystemConfig;

public class MemorySystem
{
	public static final int PAGE_OFFSET_BITS = 12;
	
	public static HashMap<String, Cache> cacheNameMappings = new HashMap<String, Cache>();
	public static HashMap<String, Coherence> coherenceNameMappings = new HashMap<String, Coherence>();
	
	public static void setCoherenceOfCaches() {
		for(Map.Entry<String, Cache> cacheListEntry :  cacheNameMappings.entrySet()) {
			Cache c = cacheListEntry.getValue();
			if(c.cacheConfig.coherenceName.equals("None")) {
				continue;
			}
			
			Coherence coherence = coherenceNameMappings.get(c.cacheConfig.coherenceName);
			c.setCoherence(coherence);
		}
	}
	
	public static Cache createSharedCache(String token, CommunicationInterface comInterface) {
		for(CacheConfig config : SystemConfig.sharedCacheConfigs) {
			if(token.equals(config.cacheName)) {
				Cache c = null;
				if(config.isDirectory==true) {
					c = new Directory(token, 0, config, null);
				}
				else if(config.nucaType != NucaType.NONE)
				{
					NucaCache nuca;
					if(!ArchitecturalComponent.nucaList.containsKey(token))
					{
						nuca = new NucaCache(token, 0, config, null);
						ArchitecturalComponent.nucaList.put(token, nuca);
					}
					else{
						nuca = ArchitecturalComponent.nucaList.get(token);
					}
					c = nuca.createBanks(token, config, comInterface);
				}
				else {
					c = new Cache(token+"[0]", 0, config, null);
				}
				c.setComInterface(comInterface);
				return c;
			}
		}
		
		misc.Error.showErrorAndExit("Unable to find a cache config for " + token);
		return null;
	}

	public static void createLinkBetweenCaches() {
		for(Map.Entry<String, Cache> cacheListEntry :  cacheNameMappings.entrySet()) {
			Cache c = cacheListEntry.getValue();
			
			// If the next level field has been set in the coreMemSys, do not set it again
			if(c.nextLevel==null) {
				createLinkToNextLevelCache(c);
			}
		}
	}
	
	private static void createLinkToNextLevelCache(Cache c) {
		String cacheName = c.cacheName;
		int cacheId = c.id;
		String nextLevelName = c.cacheConfig.nextLevel;
		
		if(nextLevelName=="" || nextLevelName==null) {
			return;
		}
		
		String nextLevelIdStrOrig = c.cacheConfig.nextLevelId;
		
		if(nextLevelIdStrOrig!=null && nextLevelIdStrOrig!="") {
			int nextLevelId = getNextLevelId(cacheName, cacheId, nextLevelIdStrOrig);
			nextLevelName += "[" + nextLevelId + "]";
		} else {
			nextLevelName += "[0]";
		}
		
		Cache nextLevelCache = cacheNameMappings.get(nextLevelName);
		if(nextLevelCache==null) {
			misc.Error.showErrorAndExit("Inside " + cacheName + ".\n" +
				"Could not find the next level cache. Name : " + nextLevelName);
		}
		
		c.createLinkToNextLevelCache(nextLevelCache);		
	}
	
	private static int getNextLevelId(String cacheName, int cacheId,
			String nextLevelIdStrOrig) {
		
		ScriptEngineManager mgr = new ScriptEngineManager();
		ScriptEngine engine = mgr.getEngineByName("JavaScript");

		//Replace $i with the component id of cache
		String cacheIdStr = (new Integer(cacheId)).toString();
		String nextLevelIdStr = nextLevelIdStrOrig.replaceAll("\\$i", cacheIdStr);
		
		try {
			Double ret =  (Double)engine.eval(nextLevelIdStr);
			return ret.intValue();
		} catch (Exception e) {
			misc.Error.showErrorAndExit("Error in evaluating the formula " +
				"for the next level cache.\n" +
				"\nname : " + cacheName + "\tid : " + cacheId + 
				"\nnextLevelIdStrOrig : " + nextLevelIdStrOrig + 
				"\nnextLevelIdStrAfterTransformation : " + nextLevelIdStr + "\n" + e);
			return -1;
		}
	}

	public static void addToCacheList(String cacheName, Cache cache) {
		if(cacheNameMappings.containsKey(cacheName)) {
			misc.Error.showErrorAndExit("A cache with same name already exists !!\nCachename : " + cacheName);
		} else {
			cacheNameMappings.put(cacheName, cache);
		}
	}
	
	public MainMemoryController getMemoryControllerId(CommunicationInterface comInterface) {
		
		MainMemoryController memControllerRet = null;
		
		if(comInterface.getClass()==NocInterface.class) {
			ID currBankId = ((NocInterface)comInterface).getId();
	    	double distance = Double.MAX_VALUE;int x1 = currBankId.getx();//bankid/cacheColumns;
	    	int y1 = currBankId.gety();//bankid%cacheColumns;
	   
	    	for(MainMemoryController memController:ArchitecturalComponent.memoryControllers) {
	    		int x2 = ((NocInterface)memController.getComInterface()).getId().getx();
	    		int y2 = ((NocInterface)memController.getComInterface()).getId().gety();
	    		double localdistance = Math.sqrt((x2-x1)*(x2-x1) + (y2-y1)*(y2-y1));
	    		if(localdistance < distance) {
	    			distance = localdistance;
	    			memControllerRet = memController;
	    		}
	    	}
		} else if(comInterface.getClass()==BusInterface.class) {
			memControllerRet = ArchitecturalComponent.memoryControllers.get(0);
		}
    	
    	return memControllerRet;
    }

}
